unit uMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls;

type
  TForm1 = class(TForm)
    Button1: TButton;
    Memo1: TMemo;
    procedure Button1Click(Sender: TObject);
  private
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}
{$MINENUMSIZE 4}

const
  IOCTL_STORAGE_QUERY_PROPERTY =  $002D1400;

type
  STORAGE_QUERY_TYPE = (PropertyStandardQuery = 0, PropertyExistsQuery, PropertyMaskQuery, PropertyQueryMaxDefined);
  TStorageQueryType = STORAGE_QUERY_TYPE;

  STORAGE_PROPERTY_ID = (StorageDeviceProperty = 0, StorageAdapterProperty);
  TStoragePropertyID = STORAGE_PROPERTY_ID;

  STORAGE_PROPERTY_QUERY = packed record
    PropertyId: STORAGE_PROPERTY_ID;
    QueryType: STORAGE_QUERY_TYPE;
    AdditionalParameters: array [0..9] of AnsiChar;
  end;
  TStoragePropertyQuery = STORAGE_PROPERTY_QUERY;

STORAGE_BUS_TYPE = (
  BusTypeUnknown = 0, BusTypeScsi = 1, BusTypeAtapi = 2, BusTypeAta = 3,
  BusType1394 = 4, BusTypeSsa = 5, BusTypeFibre = 6, BusTypeUsb = 7,
  BusTypeRAID = 8, BusTypeiScsi = 9, BusTypeSas = $A, BusTypeSata = $B,
  BusTypeSd = $C, BusTypeMmc = $D, BusTypeVirtual = $E, BusTypeFileBackedVirtual = $F,
  BusTypeMax = $10, BusTypeMaxReserved = $7F
);

//  STORAGE_BUS_TYPE = (BusTypeUnknown = 0, BusTypeScsi, BusTypeAtapi, BusTypeAta, BusType1394, BusTypeSsa, BusTypeFibre,
//    BusTypeUsb, BusTypeRAID, BusTypeiScsi, BusTypeSas, BusTypeSata, BusTypeMaxReserved = $7F);
  TStorageBusType = STORAGE_BUS_TYPE;

  STORAGE_DEVICE_DESCRIPTOR = packed record
    Version: DWORD;
    Size: DWORD;
    DeviceType: Byte;
    DeviceTypeModifier: Byte;
    RemovableMedia: Boolean;
    CommandQueueing: Boolean;
    VendorIdOffset: DWORD;
    ProductIdOffset: DWORD;
    ProductRevisionOffset: DWORD;
    SerialNumberOffset: DWORD;
    BusType: STORAGE_BUS_TYPE;
    RawPropertiesLength: DWORD;
    RawDeviceProperties: array [0..0] of AnsiChar;
  end;
  TStorageDeviceDescriptor = STORAGE_DEVICE_DESCRIPTOR;

function GetBusType(Drive: AnsiChar): TStorageBusType;
var
  H: THandle;
  Query: TStoragePropertyQuery;
  dwBytesReturned: DWORD;
  Buffer: array [0..1023] of Byte;
  sdd: TStorageDeviceDescriptor absolute Buffer;
  OldMode: UINT;
begin
  Result := BusTypeUnknown;

  OldMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  try
    H := CreateFile(PChar(Format('\\.\%s:', [AnsiLowerCase(Drive)])), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
      OPEN_EXISTING, 0, 0);
    if H <> INVALID_HANDLE_VALUE then
    begin
      try
        dwBytesReturned := 0;
        FillChar(Query, SizeOf(Query), 0);
        FillChar(Buffer, SizeOf(Buffer), 0);
        sdd.Size := SizeOf(Buffer);
        Query.PropertyId := StorageDeviceProperty;
        Query.QueryType := PropertyStandardQuery;
        if DeviceIoControl(H, IOCTL_STORAGE_QUERY_PROPERTY, @Query, SizeOf(Query), @Buffer, SizeOf(Buffer), dwBytesReturned, nil) then
          Result := sdd.BusType;
      finally
        CloseHandle(H);
      end;
    end;
  finally
    SetErrorMode(OldMode);
  end;
end;

function GetPhysicalDiskNumber(Drive: Char): Cardinal;

  function GetLD(Drive: Char): Cardinal;
    var
      Buffer : String;
    begin
      Buffer := Format('\\.\%s:',[Drive]);
      Result := CreateFile(PChar(Buffer), GENERIC_READ,
                           FILE_SHARE_READ or FILE_SHARE_WRITE,
                           nil, OPEN_EXISTING, 0, 0);
      Win32Check(Result <> INVALID_HANDLE_VALUE);
    end;

type
  PDiskInfo = ^TDiskInfo;
  TDiskInfo = record
    BootStatus,
    StartHead    : Byte;
    StartSecClu  : Array[0..1]  Of Byte;
    ParitionType,
    LastHead     : Byte;
    LastSecClu   : Array[0..1]  Of Byte;
    ABSSector,
    TTLSector    : Integer;
    Reserved     : Array[0..47] Of Byte;
    Signature    : Array[0..1]  Of Byte;
  end;
  TDiskExtent = record
    DiskNumber: Cardinal;
    StartingOffset: Int64;
    ExtentLength: Int64;
  end;
  DISK_EXTENT = TDiskExtent;
  PDiskExtent = ^TDiskExtent;
  TVolumeDiskExtents = record
    NumberOfDiskExtents: Cardinal;
    Extents: array[0..0] of TDiskExtent;
  end;
  VOLUME_DISK_EXTENTS = TVolumeDiskExtents;
  PVolumeDiskExtents = ^TVolumeDiskExtents;

const
  FILE_DEVICE_DISK                     = $00000007;
  METHOD_BUFFERED                      = 0;
  FILE_ANY_ACCESS                      = 0;
  IOCTL_DISK_BASE                      = FILE_DEVICE_DISK;
  IOCTL_VOLUME_BASE                    = DWORD('V');
  IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS = ((IOCTL_VOLUME_BASE shl 16) or (FILE_ANY_ACCESS shl 14) or (0 shl 2) or METHOD_BUFFERED);

var
  LD : DWORD;
  DiskExtents : PVolumeDiskExtents;
  DiskExtent : TDiskExtent;
  BytesReturned : Cardinal;
begin
  Result := 255;
  LD := GetLD(Drive);
  If LD = INVALID_HANDLE_VALUE Then Exit;
    Try
      DiskExtents := AllocMem(Max_Path);
      try
        DeviceIOControl(LD,IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,nil,0,DiskExtents,Max_Path,BytesReturned,nil);
        If DiskExtents^.NumberOfDiskExtents > 0 Then
        begin
          DiskExtent := DiskExtents^.Extents[0];
          Result := DiskExtent.DiskNumber;
        end;
      finally
        FreeMem(DiskExtents);
      end;
    Finally
      CloseHandle(LD);
    end;
end;

FUNCTION DiskInDrive(CONST Drive: CHAR): BOOLEAN;
  VAR
    DriveNumber:  BYTE;
    ErrorMode  : Word;
BEGIN
  RESULT := FALSE;
  DriveNumber := ORD( UpCase(Drive) );
  ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  TRY
    // 'A'=1, 'B'=2, ... in DiskSize call
    IF   DiskSize(DriveNumber-ORD('A')+1) <> -1
    THEN RESULT := TRUE
  FINALLY
    SetErrorMode(ErrorMode)
  end;
END;


FUNCTION IsDiskWriteProtected(CONST drive:  CHAR):  BOOLEAN;
  VAR
    ErrorMode:  Word;
    PathName :  STRING;
    TempName :  STRING;
BEGIN
  ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  TRY
    ASSERT (Upcase(drive) IN ['A'..'Z'], 'Invalid drive specification');
    PathName := drive + ':\';
    SetLength(TempName, MAX_PATH+1);
    GetTempFileName(pChar(PathName), 'RWRO', 0, pChar(TempName));
    RESULT := (GetLastError = Windows.ERROR_WRITE_PROTECT);
    IF NOT RESULT
    THEN BEGIN
      RESULT := NOT DeleteFile(TempName);
    END
  FINALLY
    SetErrorMode(ErrorMode)
  END
END;

PROCEDURE ShowDrives(CONST MemoInfo:  TMemo);
  VAR
    DriveBits         :  SET OF 0..25;
    DriveIndex        :  INTEGER;
    DriveLetter       :  CHAR;
    DriveType         :  Cardinal;
    DriveString       :  String;
    NotUsed           :  DWORD;
    ReadWrite         :  STRING;
    s                 :  STRING;
    VolumeFlags       :  DWORD;
    VolumeInfo        :  ARRAY[0..MAX_PATH] OF CHAR;
    VolumeSerialNumber:  INTEGER;
    PhysicalDisk      :  Byte;

function GetDriveTypeString(DriveByte: Byte): String;
begin
  case GetDriveType(PChar(Chr(DriveByte + Ord('A')) + ':\')) of
    DRIVE_UNKNOWN: Result := ('Unknown');
    DRIVE_NO_ROOT_DIR: Result := ('not Mounted');
    DRIVE_REMOVABLE: Result := ('Removable');
    DRIVE_FIXED: Result := ('Fixed');
    DRIVE_REMOTE: Result := ('Network');
    DRIVE_CDROM: Result := ('CDROM');
    DRIVE_RAMDISK: Result := ('RAM');
  else
    Result := 'Other';
  end;
end;

  FUNCTION FormatBytes(CONST Bytes:  Int64):  STRING;
  BEGIN
    IF   Bytes < 0
    THEN RESULT := 'not available'
    ELSE RESULT := FormatFloat('0,', Bytes)
  END;

begin
  Screen.Cursor := crHourGlass;
  TRY
    MemoInfo.Lines.Add('');
    MemoInfo.Lines.Add(' Drv Letter   Type@Bus R/W          Total Bytes           Free Bytes Vol-Ser# Label');
    MemoInfo.Lines.Add('---------------------- --- -------------------- -------------------- -------- ------------------>');
    Cardinal(DriveBits) := GetLogicalDrives;
    FOR  DriveIndex := 0 TO 25 DO
    BEGIN
      IF  DriveIndex IN DriveBits
      THEN BEGIN
        DriveLetter := CHR(ORD('A') + DriveIndex);
        PhysicalDisk := GetPhysicalDiskNumber(DriveLetter);
        DriveType   := GetDriveType(pChar(DriveLetter + ':\'));
        GetVolumeInformation(pChar(DriveLetter + ':\'),
                             VolumeInfo, SizeOf(VolumeInfo),
                             @VolumeSerialNumber, NotUsed, VolumeFlags, NIL, 0);
        DriveString := GetDriveTypeString(DriveIndex)+'@';
        case GetBusType(AnsiChar(DriveLetter)) of
          BusTypeUnknown: DriveString := DriveString + ('Unknown');
          BusTypeScsi: DriveString := DriveString + ('Scsi');
          BusTypeAtapi: DriveString := DriveString + ('Atapi');
          BusTypeAta: DriveString := DriveString + ('Ata');
          BusType1394: DriveString := DriveString + ('1394');
          BusTypeSsa: DriveString := DriveString + ('Ssa');
          BusTypeFibre: DriveString := DriveString + ('Fibre');
          BusTypeUsb: DriveString := DriveString + ('Usb');
          BusTypeRAID: DriveString := DriveString + ('RAID');
          BusTypeiScsi: DriveString := DriveString + ('iScsi');
          BusTypeSas: DriveString := DriveString + ('Sas');
          BusTypeSata: DriveString := DriveString + ('Sata');
          BusTypeSd: DriveString := DriveString + ('Sd');
          BusTypeMmc: DriveString := DriveString + ('Mmc');
          BusTypeVirtual: DriveString := DriveString + ('Virtual');
          BusTypeFileBackedVirtual: DriveString := DriveString + ('FileBackedVirtual');
          BusTypeMax: DriveString := DriveString + ('Max');
          BusTypeMaxReserved: DriveString := DriveString + ('MaxReserved');
        else
          DriveString := DriveString + ('Other');
        end; // case
        IF DiskInDrive(DriveLetter)
        THEN BEGIN
          IF IsDiskWriteProtected(DriveLetter)
          THEN ReadWrite := 'R/O'
          ELSE ReadWrite := 'R/W';
          s := Format(' %3u %s %15s %3s %20s %20s %8.8X %s',
                      [PhysicalDisk,
                       DriveLetter,
                       DriveString,
                       ReadWrite,
                       FormatBytes(DiskSize(DriveIndex+1)),
                       FormatBytes(DiskFree(DriveIndex+1)),
                       VolumeSerialNumber,
                       VolumeInfo
                      ]
                     )
        END
        ELSE  s := Format('  %s %-7s %20s',
                      [DriveLetter,
                       GetDriveTypeString(DriveType),
                       '<disk missing>']);
        MemoInfo.Lines.Add(s);
      END
    END;

    MemoInfo.Lines.Add(' ')
  FINALLY
    Screen.Cursor := crDefault
  END
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  Screen.Cursor := crHourGlass;
  Memo1.Clear;
  Memo1.Lines.Add('KodeZwerg was here ;-)');
  Memo1.Lines.Add('working... (if nothing happens my App failed, i am sorry!)');
  Memo1.Lines.BeginUpdate;
  TRY
    ShowDrives(Memo1)
  FINALLY
    Screen.Cursor := crDefault
  END;
  Memo1.Lines.EndUpdate;
end;

end.
